/* $Id: multip.c,v 1.5.3 2004/07/24 08:50:06 andres Exp $ */
/*
 *	Stonith module for Multip NISPERO Series Power Switch
 *
 *      Original code from baytech.c by
 *	Copyright (c) 2000 Alan Robertson <alanr@unix.sh>
 *
 *      Modifications for NW RPC100S
 *	Copyright (c) 2000 Computer Generation Incorporated
 *               Eric Z. Ayers <eric.ayers@compgen.com>
 *	
 *	Modifications for Multip NISPERO Series
 *	Copyright (c) 2004 Mission Critical Linux Kft
 *	         Andres Sajo <andres@mclx.hu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *
 *	Note for the Multip NISPERO Series:
 *	
 *	* Reset is handled by off -> sleep(ctx->reset_time) -> on
 *	* if reset_time is 0, then leave the outlet(s) OFF.
 *	
 */

#include <portability.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <syslog.h>
#include <libintl.h>
#include <sys/wait.h>
#include <termios.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <glib.h>

/* Unix Domain Socket */
#include <sys/socket.h>
#include <sys/un.h>


#include <stonith/stonith.h>
#define PIL_PLUGINTYPE          STONITH_TYPE
#define PIL_PLUGINTYPE_S        STONITH_TYPE_S
#define PIL_PLUGIN              m_nispero
#define PIL_PLUGIN_S            "m_nispero"
#define PIL_PLUGINLICENSE 	LICENSE_LGPL
#define PIL_PLUGINLICENSEURL 	URL_LGPL
#include <pils/plugin.h>

/* these are to be unset before production version: */
#define DEBUG 3
#define SECURITY_LOW

#ifndef DEBUG
#define DEBUG 0
#endif
static int gbl_debug = DEBUG;

#define RESET_INT 10 /* seconds */

/* Policy defines */
/*These should go in the config file*/
#define N_USER (const char *)"REGISTER:STONITH"
/*These should go in the config file*/
#define N_PASSWORD (const char *)"PASSWORD:"
#define N_STATUS (const char *)"STATUS:"
#define N_ON (const char *)"ON:"
#define N_OFF (const char *)"OFF:"

/* Answers from daemon */
#define N_OK "OK"
#define N_INVALID "INVALID"


#define	DEVICE	"Multip NISPERO Series Power Switch"
#define N_(text)	(text)
#define _(text)		dgettext(ST_TEXTDOMAIN, text)

/*
 * multip_nisperoclosepi is called as part of unloading the multip_nispero
 * STONITH plugin.
 * If there was any global data allocated, or file descriptors opened, etc.
 * which is associated with the plugin, and not a single interface
 * in particular, here's our chance to clean it up.
 */

static void
multip_nisperoclosepi(PILPlugin*pi)
{
}


/*
 * multip_nisperocloseintf called as part of shutting down the multip_nispero
 * STONITH interface. 
 * If there was any global data allocated, or file descriptors
 * opened, etc.  which is associated with the multip_nispero implementation,
 * here's our chance to clean it up.
 */
static PIL_rc
multip_nisperocloseintf(PILInterface* pi, void* pd)
{
	return PIL_OK;
}

static void *	multip_nispero_new(void);
static void	multip_nispero_destroy(Stonith *);
static int	multip_nispero_set_config_file(Stonith *, const char * cfgname);
static int	multip_nispero_set_config_info(Stonith *, const char * info);
static const char *	multip_nispero_getinfo(Stonith * s, int InfoType);
static int	multip_nispero_status(Stonith * );
static int	multip_nispero_reset_req(Stonith * s, int request, const char * host);
static char **	multip_nispero_hostlist(Stonith  *);
static void	multip_nispero_free_hostlist(char **);

static struct stonith_ops multip_nisperoOps ={
	multip_nispero_new,		/* Create new STONITH object	*/
	multip_nispero_destroy,		/* Destroy STONITH object	*/
	multip_nispero_set_config_file,	/* set configuration from file	*/
	multip_nispero_set_config_info,	/* Get configuration from file	*/
	multip_nispero_getinfo,		/* Return STONITH info string	*/
	multip_nispero_status,		/* Return STONITH device status	*/
	multip_nispero_reset_req,	/* Request a reset */
	multip_nispero_hostlist,	/* Return list of supported hosts */
	multip_nispero_free_hostlist	/* free above list */
};

PIL_PLUGIN_BOILERPLATE("1.0", Debug, multip_nisperoclosepi);
static const PILPluginImports*  PluginImports;
static PILPlugin*               OurPlugin;
static PILInterface*		OurInterface;
static StonithImports*		OurImports;
static void*			interfprivate;

#define LOG		PluginImports->log
#define MALLOC		PluginImports->alloc
#define STRDUP  	PluginImports->mstrdup
#define FREE		PluginImports->mfree
#define EXPECT_TOK	OurImports->ExpectToken
#define STARTPROC	OurImports->StartProcess

PIL_rc
PIL_PLUGIN_INIT(PILPlugin*us, const PILPluginImports* imports);

PIL_rc
PIL_PLUGIN_INIT(PILPlugin*us, const PILPluginImports* imports)
{
	/* Force the compiler to do a little type checking */
	(void)(PILPluginInitFun)PIL_PLUGIN_INIT;

	PluginImports = imports;
	OurPlugin = us;

	/* Register ourself as a plugin */
	imports->register_plugin(us, &OurPIExports);  

	/*  Register our interface implementation */
 	return imports->register_interface(us, PIL_PLUGINTYPE_S
	,	PIL_PLUGIN_S
	,	&multip_nisperoOps
	,	multip_nisperocloseintf		/*close */
	,	&OurInterface
	,	(void*)&OurImports
	,	&interfprivate); 
}

/* 

   The Multip NISPERO Series Power Switch is manufatured by:
   
	Multip Kft
	++36 30 382 92 86
	http://www.multip.hu/
	staff@multip.hu
	
   Mission Critical Linux is the mantainer of this plugin.
   http://www.mclx.hu/andres
   Andres Sajo <andres@mclx.hu>
   
   A simple view of the architecture of this unit ...
   
           | in| | in| | in| | in| | in| | in|               <ac in/out lets>
           |out| |out| |out| |out| |out| |out|
         +-------------------------------------+
         |              REELAY ARRAY           |
         |               2 to 6 pcs            |
         +----------------+-----+--------------+
                          |     |
                          ^     ^
   +---+----/ 6 bit /-->--+-----+--<--/ 6 bit /-+-----+     <6 bit bus>
   | 6 |                  |  6  |               |  6  |
  +-----+                 +-----+               +-----+     <1wire ics
  |2480 |                 |2480 |               |2480 |      Dallas Instruments
  +-----+                 +-----+               +-----+      2480>
     |                       |                     |
     +-----------------------+---------------------+        <serial to 1wire bus>
                             |
                   +------------------+
                   |   SERIAL DEVICE  |
                   +------------------+
                            ^
                            |
                            v
   +-------------------------------------------------+
   |                  1WIRE API                      |
   +-------------------------------------------------+
   |                  NISPERO DAEMON                 |     <see protocol description>
   +-------------------------------------------------+
        ^                   ^
        |                   |
        |                   |                              <unix domain sockets>
        v                   v
   +-----------+       +----------+
   | heartbeat |       | nispero  |
   |  stonith  |       | utility  |
   +-----------+       +----------+  
   

   nisperod:
   nisperod is the controller of the power switch, and it's protocol is fairly simple:
   "REGISTER:STONITH\n"     <--- don't change this ! (check note 1)
   "PASSWORD:<PASSWORD>\n"
   "STATUS:<host>\n"
   "ON:<host>\n"
   "OFF:<host>\n"
   
   
   A response is either
   "OK\n"
   or
   "INVALID\n"
   
   NOTES:     
   
   1)
   The daemon has two type of protocoll: simple (aka stonith) and extended.
   This plugin communicates with the simple protocol.
   For more information on the nispero protocols check out the protocols.readme file
   
   2)
   The daemon won't let you operate the switch without registering username (first) and password.
   
*/
/* Template :
struct NW_RPC100S { 
	const char *	RPCid;
	char *	idinfo;  ??? What's this for Alan ???
	char *	unitid;  ??? What's this for Alan ???
	int	fd;      FD open to the serial port
	int	config;  0 if not configured, 
			 1 if configured with multip_nispero_set_config_file() 
			 or multip_nispero_set_config_info()
	char *	device;  Serial device name to use to communicate 
                         to this RPS10
	char *  node;    Name of the node that this is controlling
};
*/

struct M_NISPERO {
	const char *	RPCid;
	char *		idinfo;
	char *		unitid;
	int 		fd;	/* socket */ 
	int 		config;
	char *		device;	/* filename of the socket */
	char *		node;	/* name of the node that this is controlling */
	char *		password; /* password to authentificate to the daemon */
	int 		reset_time; /* time between OFF -> ON, 0 (zero) to leave 
	                               outlet(s) OFF, this means manual intervention to
				       restore power. */
};

/* This string is used to identify this type of object in the config file */
static const char * RPCid = "M_NISPERO";
static const char * NOTrpcid = "OBJECT DESTROYED: (M_NISPERO)";



#define	ISM_NISPERO(i)	(((i)!= NULL && (i)->pinfo != NULL)	\
	&& ((struct M_NISPERO *)(i->pinfo))->RPCid == RPCid)

#define	ISCONFIGED(i)	(ISM_NISPERO(i) && ((struct M_NISPERO *)(i->pinfo))->config)

#ifndef MALLOCT
#	define     MALLOCT(t)      ((t *)(MALLOC(sizeof(t)))) 
#endif

#define WHITESPACE	" \t\n\r\f"

#define	REPLSTR(s,v)	{					\
			if ((s) != NULL) {			\
				FREE(s);			\
				(s)=NULL;			\
			}					\
			(s) = STRDUP(v);			\
			if ((s) == NULL) {			\
				syslog(LOG_ERR, _("out of memory"));\
			}					\
			}

static int	RPCConnect(struct M_NISPERO * ctx);
static int	RPCDisconnect(struct M_NISPERO * ctx);
/* The following functions don't handle RPC{Connect|Disconnect} 
 * they are handled by their callers
 */
static int	RPCReset(struct M_NISPERO*, const char * host);
static int	RPCOn(struct M_NISPERO*, const char * host);
static int	RPCOff(struct M_NISPERO*, const char * host);
static int	RPC_parse_config_info(struct M_NISPERO* ctx, const char * info);

/* 
 * RPCReset - Reset (power-cycle) the given host
 */
static int
RPCReset(struct M_NISPERO* ctx, const char * host)
{
	if (gbl_debug) syslog(LOG_DEBUG, "%s: calling RPCReset host=%s\n",
	RPCid, host);

	if ( RPCOff(ctx, host) != S_OK ) {
		return S_OOPS;
	}

	if (ctx->reset_time == 0) {
		if (gbl_debug) syslog(LOG_DEBUG, "%s: leaving %s OFF", RPCid, host);
		return(S_OK);
	}
	if (gbl_debug) syslog(LOG_DEBUG, "%s: sleeping %i", RPCid, ctx->reset_time);
	if (gbl_debug) syslog(LOG_DEBUG, "%s: sleep = %i", RPCid, sleep(ctx->reset_time));
	
	if (RPCOn(ctx, host)){
		return S_OOPS;
	}
	return(S_OK);

} /* end RPCReset() */


/* 
 * RPCStatus - Get status of the given unit 
 */
static int
RPCStatus(struct M_NISPERO* ctx)
{
	char buf[1024];
	int amount;

	if (gbl_debug) syslog(LOG_DEBUG, "%s: calling RPCStatus", RPCid);
	if (ctx->fd < 0) {
		syslog(LOG_ERR, "%s: socket %s is not open!", RPCid, 
		       ctx->device);
		return S_OOPS;
	}
	
	/* send the "STATUS:" command and expect N_OK */
	strcpy(buf, N_STATUS);
	strcat(buf, ctx->node);
	if (gbl_debug) syslog(LOG_DEBUG, "%s: sending string \'%s\' + \\n", RPCid, buf);
	strcat(buf, (const char *)"\n");	
	amount = strlen(buf);
	if (gbl_debug) syslog(LOG_DEBUG, "%s: sending string len = %i", RPCid, strlen(buf));
	if (write(ctx->fd, buf, amount) != amount) 
	{
		syslog (LOG_ERR, "%s: can't write to socket %s : %s",
		RPCid, ctx->device, strerror(errno));
		return S_OOPS;
	}
	amount = read(ctx->fd, buf, sizeof(buf));
	if (amount <= 0){
		syslog (LOG_ERR, "%s: can't read from socket %s : %s",
		RPCid, ctx->device, strerror(errno));
		return S_OOPS;
	}
	if ( strncmp(buf, N_OK, strlen(N_OK)) != 0){
		syslog (LOG_ERR, "%s: can't get the status of the device",
		RPCid);
		return S_OOPS;
	}
	return S_OK;
} /* end RPCStatus() */


/*
 * RPCOn - Turn On the given outlet number 
 */
static int
RPCOn(struct M_NISPERO* ctx, const char * host)
{
	char buf[1024];
	int amount;
	
	if (gbl_debug) syslog(LOG_DEBUG, "%s: calling RPCOn host=%s\n",
	RPCid, host);
	
	if (ctx->fd < 0) {
		syslog(LOG_ERR, "%s: socket %s is not open!", RPCid, 
		       ctx->device);
		return S_OOPS;
	}
	
	/* send the "On" command */
	strcpy(buf, N_ON);
	strcat(buf, host);
	if (gbl_debug) syslog(LOG_DEBUG, "%s: calling RPCOn with '%s' + \\n", RPCid, buf);
	strcat(buf, (const char *)"\n");
	amount = strlen(buf);
	if (write(ctx->fd, buf, amount) != amount) 
	{
		syslog (LOG_ERR, "%s: can't write to socket %s : %s",
		RPCid, ctx->device, strerror(errno));
		return S_OOPS;
	}
	amount = read(ctx->fd, buf, sizeof(buf));
	if (amount <= 0){
		syslog (LOG_ERR, "%s: can't read from socket %s : %s",
		RPCid, ctx->device, strerror(errno));
		return S_OOPS;
	}
	if ( strncmp(buf, N_OK, strlen(N_OK)) != 0){
		syslog (LOG_ERR, "%s: error turning on the outlet(s) of host %s",
		RPCid, host);
		return S_OOPS;
	}
	return S_OK;	

} /* end RPCOn() */


/*
 * RPCOff - Turn Off the given outlet number 
 */
static int
RPCOff(struct M_NISPERO* ctx, const char * host)
{
	char buf[1024];
	int amount;
	
	if (gbl_debug) syslog(LOG_DEBUG, "%s: calling RPCOff host=%s\n",
	RPCid, host);
	if (ctx->fd < 0) {
		syslog(LOG_ERR, "%s: socket %s is not open!", RPCid, 
		       ctx->device);
		return S_OOPS;
	}

	/* send the "Off" command */
	strcpy(buf, N_OFF);
	strcat(buf, host);
	if (gbl_debug) syslog(LOG_DEBUG, "%s: calling RPCOff with '%s' + \\n", RPCid, buf);
	strcat(buf, (const char *)"\n");

	amount = strlen(buf);
	if (write(ctx->fd, buf, amount) != amount) 
	{
		syslog (LOG_ERR, "%s: can't write to socket %s : %s",
		RPCid, ctx->device, strerror(errno));
		RPCDisconnect(ctx);
		return S_OOPS;
	}
	amount = read(ctx->fd, buf, sizeof(buf) );
	if (amount <= 0){
		syslog (LOG_ERR, "%s: can't read from socket %s : %s",
		RPCid, ctx->device, strerror(errno));
		return S_OOPS;
	}
	if ( strncmp(buf, N_OK, strlen(N_OK)) != 0){
		syslog (LOG_ERR, "%s: error turning off the outlet(s) of host %s",
		RPCid, host);
		return S_OOPS;
	}
	return S_OK;	
} /* end RPCOff() */


/*
 * multip_nispero_status - API entry point to probe the status of the stonith device 
 *           (basically just "is it reachable and functional?", not the
 *            status of the individual outlets)
 * 
 * Returns:
 *    S_OOPS - some error occured
 *    S_OK   - if the stonith device is reachable and online.
 */
static int
multip_nispero_status(Stonith  *s)
{
	struct M_NISPERO*	ctx;
	
	if (gbl_debug) syslog(LOG_DEBUG, "%s: calling multip_nispero_status", RPCid);
	
	if (!ISM_NISPERO(s)) {
		syslog(LOG_ERR, "%s: invalid argument to multip_nispero_status", RPCid);
		return(S_OOPS);
	}
	if (!ISCONFIGED(s)) {
		syslog(LOG_ERR
		,	"%s: unconfigured stonith object in multip_nispero_status", RPCid);
		return(S_OOPS);
	}
	ctx = (struct M_NISPERO*) s->pinfo;
	if (RPCConnect(ctx) != S_OK) {
		return(S_OOPS);
	}
	if (RPCStatus(ctx) != S_OK) {
		RPCDisconnect(ctx);
		return(S_OOPS);
	}
	RPCDisconnect(ctx);
	return(S_OK);
}

/*
 * multip_nispero_hostlist - API entry point to return the list of hosts 
 *                 for the devices on this M_NISPERO unit
 * 
 *               This type of device is configured from the config file,
 *                 so we don't actually have to connect to figure this
 *                 out, just peruse the 'ctx' structure.
 * Returns:
 *     NULL on error
 *     a malloced array, terminated with a NULL,
 *        of null-terminated malloc'ed strings.
 */
static char **
multip_nispero_hostlist(Stonith  *s)
{
	char **	ret = NULL;	/* list to return */
	struct M_NISPERO*	ctx;
	
	if (gbl_debug) syslog(LOG_DEBUG, "%s: calling multip_nispero_hostlist", RPCid);
	
	if (!ISM_NISPERO(s)) {
		syslog(LOG_ERR, "%s: invalid argument to RPC_list_hosts", RPCid);
		return(NULL);
	}
	if (!ISCONFIGED(s)) {
		syslog(LOG_ERR
		,	"%s: unconfigured stonith object in RPC_list_hosts", RPCid);
		return(NULL);
	}

	ctx = (struct M_NISPERO*) s->pinfo;

	ret = (char **)MALLOC(2*sizeof(char*));
	if (ret == NULL) {
		syslog(LOG_ERR, "%s: out of memory", RPCid);
	} else {
		ret[1]=NULL;
		ret[0]=STRDUP(ctx->node);
		if (ret[0] == NULL) {
			syslog(LOG_ERR, "%s: out of memory", RPCid);
			FREE(ret);
			ret = NULL;
		} else {
			g_strdown(ret[0]);
		}
	}

	return(ret);
} /* end si_hostlist() */

/*
 * multip_nispero_free_hostlist - free the result from multip_nispero_hostlist() 
 */
static void
multip_nispero_free_hostlist (char ** hlist)
{
	char **	hl = hlist;
	if (gbl_debug) syslog(LOG_DEBUG, "%s: calling multip_nispero_free_hostlist", RPCid);
	if (hl == NULL) {
		return;
	}
	while (*hl) {
		FREE(*hl);
		*hl = NULL;
		++hl;
	}
	FREE(hlist);
}


/* TODO: TO_TEST FOR MORE THAN TWO NODES
 *      Parse the given configuration information, and stash it away...
 *
 *      <info> contains the parameters specific to this type of object
 *
 *         The format of <parameters> for this module is:
 *            <socket name> <owner_node> <password>
 *
 *      e.g. A machine named 'nodea' can kill a machine named 'nodeb' through
 *           a socket /tmp/socket.
 *           A machine named 'nodeb' can kill machines 'nodea' and 'nodec'
 *           through a socket named /tmp/socket
 *
 *      stonith M_NISPERO /tmp/socket nodea ...
 *      stonith M_NISPERO /tmp/socket nodeb ...
 *      stonith M_NISPERO /tmp/socket nodec ...
 *
 *
 */

static int
RPC_parse_config_info(struct M_NISPERO* ctx, const char * info)
{
	char *copy;
	char *token;
	if (gbl_debug) syslog(LOG_DEBUG, "%s: calling RPC_parse_config_info", RPCid);
	if (ctx->config) {
		/* The module is already configured. */
		return(S_OOPS);
	}

	/* strtok() is nice to use to parse a string with 
	   (other than it isn't threadsafe), but it is destructive, so
	   we're going to alloc our own private little copy for the
	   duration of this function.
	*/

	copy = STRDUP(info);
	if (!copy) {
		syslog(LOG_ERR, "%s: out of memory", RPCid);
		return S_OOPS;
	}

	/* Grab the serial device */
	token = strtok (copy, " \t");
	if (!token) {
		/* syslog(LOG_ERR, "%s: Can't find serial device on config line '%s'",*/
		syslog(LOG_ERR, "%s: can't find socket path in config line '%s'",
		       RPCid, info);
		goto token_error;		
	}

	ctx->device = STRDUP(token);
	if (!ctx->device) {
		syslog(LOG_ERR, "%s: out of memory", RPCid);
		goto token_error;
	}


	/* 
	 * Grab <nodename>
	 */
	 
	token = strtok (NULL, " \t");
	if (!token) {
		syslog(LOG_ERR, "%s: can't find node name on config line '%s'",
		       RPCid, info);
		goto token_error;		
	}

	ctx->node = STRDUP(token);
	if (!ctx->node) {
		syslog(LOG_ERR, "%s: out of memory", RPCid);
		goto token_error;
	}
	
	/* 
	 * Grab <password> for STONITH user
	 */
	 
	token = strtok (NULL, " \t");
	if (!token) {
		syslog(LOG_ERR, "%s: can't find password on config line '%s'",
		       RPCid, info);
		goto token_error;		
	}

	ctx->password = STRDUP(token);
	if (!ctx->password) {
		syslog(LOG_ERR, "%s: out of memory", RPCid);
		goto token_error;
	}
	
	/* grab <reset_time> */

	token = strtok (NULL, " \t");
	if (!token) {
		syslog(LOG_ERR, "%s: can't find reset_time on config line '%s'", RPCid, info);
		goto token_error;
	}
	ctx->reset_time = atoi(token);
	if (ctx->reset_time > 3600){
		syslog(LOG_ERR, "%s: reset_time > 3600, you might as well set it to 0", RPCid);
		goto token_error;
	}
	if (ctx->reset_time < 0){
		syslog(LOG_ERR, "%s: reset_time out of range", RPCid);
		goto token_error;
	}
		
	/* free our private copy of the string we've been destructively 
	   parsing with strtok()
	*/
	FREE(copy);
	ctx->config = 1;
#ifdef SECURITY_LOW
	if (gbl_debug) syslog(LOG_DEBUG, "%s: parse_cnf_info: RPCid='%s', "
	"idinfo='%s', unitid='%s', device='%s', node='%s', password='%s', reset_time=%i",
	RPCid, ctx->RPCid, ctx->idinfo, ctx->unitid, ctx->device, ctx->node,
	ctx->password, ctx->reset_time);
#endif
	return S_OK;

token_error:
	FREE(copy);
	return(S_BADCONFIG);
}

/* 
 * RPCConnect
 *
 * Connect to the Nispero Daemon.
 * Side Effects
 *  1 ctx->fd now contains a valid file descriptor to an open socket
 *  2 calling twice this function, returns S_OOPS, call RPCDisconnect
 *    before recalling this function.
 *  
 * Returns 
 *    S_OK on success
 *    S_OOPS on error
 *
 */
static int
RPCConnect(struct M_NISPERO * ctx)
{
	struct sockaddr_un address;
	size_t addrLength;
	char buf[1024];
	int amount;
 	if (gbl_debug) syslog(LOG_DEBUG, "%s: calling RPCConnect", RPCid);
	
	/* Open the socket if it isn't already open */
	if (ctx->fd < 0) {
		if((ctx->fd = socket(PF_UNIX, SOCK_STREAM, 0)) < 0)
		{
			syslog (LOG_ERR, "%s: can't open socket %s : %s",
			RPCid, ctx->device, strerror(errno));
			return S_OOPS;
		}
		address.sun_family = AF_UNIX;
		strcpy(address.sun_path, ctx->device);
		addrLength = sizeof(address.sun_family) + strlen(address.sun_path);
		if(connect(ctx->fd, (struct sockaddr *) &address, addrLength))
		{
			syslog (LOG_ERR, "%s: can't open socket %s : %s",
			RPCid, ctx->device, strerror(errno));
			return S_OOPS;
		}
	} else {
		syslog (LOG_ERR, "%s: trying to Connect to an already "
		"opened socket, this shouln't be, close socket first.",
		RPCid);
		return S_OOPS;
	}
	syslog (LOG_ERR, "%s: registering device",
		RPCid);
	/* Register the device: */
	strcpy(buf, N_USER);
	strcat(buf, (const char *)"\n");	
	amount = strlen(buf);
	if (write(ctx->fd, buf, amount) != amount) 
	{
		syslog (LOG_ERR, "%s: can't write to socket %s : %s",
		RPCid, ctx->device, strerror(errno));
		return S_OOPS;
	}
	amount = read(ctx->fd, buf, sizeof(buf));
	if ((amount) <= 0){
		syslog (LOG_ERR, "%s: can't read from socket %s : %s",
		RPCid, ctx->device, strerror(errno));
		return S_OOPS;
	}
	if (strncmp(buf, N_OK, strlen(N_OK)) != 0){
		syslog (LOG_ERR, "%s: can't login to daemon",
		RPCid);
		return S_OOPS;
	}
	/* we are registered ! */
	syslog (LOG_ERR, "%s: authentificating device",
		RPCid);
	strcpy(buf, N_PASSWORD);
	strcat(buf, ctx->password);
	strcat(buf, (const char *)"\n");
	amount = strlen(buf);
	if (write(ctx->fd, buf, amount) != amount) 
	{
		syslog (LOG_ERR, "%s: can't write to socket %s : %s",
		RPCid, ctx->device, strerror(errno));
		return S_OOPS;
	}
	amount = read(ctx->fd, buf, sizeof(buf));
	if ((amount)<= 0){
		syslog (LOG_ERR, "%s: can't read from socket %s : %s",
		RPCid, ctx->device, strerror(errno));
		return S_OOPS;
	}
	if ( strncmp(buf, N_OK, strlen(N_OK)) != 0){
		syslog (LOG_ERR, "%s: can't validate from daemon",
		RPCid);
		return S_OOPS;
	}
	
	return(S_OK);
}

static int
RPCDisconnect(struct M_NISPERO * ctx)
{
	if (gbl_debug) syslog(LOG_DEBUG, "%s: calling RPCDisconnect", RPCid);
	if (ctx->fd >= 0) {
		/* Flush the socket, we don't care what happens to the characters
		 * and failing to do this can cause close to hang.
		 */
		/* tcflush(ctx->fd, TCIOFLUSH); */
		/* TODO: sockets need to be flushed ???*/
		close (ctx->fd);
	}
	ctx->fd = -1;
	return S_OK;
}

/*
 *	multip_nispero_reset - API call to Reset (reboot) the given host on 
 *          this Stonith device.  This involves toggling the power off 
 *          and then on again, OR just calling the builtin reset command
 *          on the stonith device.
 */
static int
multip_nispero_reset_req(Stonith * s, int request, const char * host)
{
	int	rc = S_OK;
	int	lorc = S_OK;
	struct M_NISPERO*	ctx;
	
	if (gbl_debug) syslog(LOG_DEBUG, "%s: calling multip_nispero_reset", RPCid);
	
	if (!ISM_NISPERO(s)) {
		syslog(LOG_ERR, "%s: invalid argument to RPC_reset_host", RPCid);
		return(S_OOPS);
	}
	if (!ISCONFIGED(s)) {
		syslog(LOG_ERR
		,	"%s: unconfigured stonith object in RPC_reset_host", RPCid);
		return(S_OOPS);
	}
	ctx = (struct M_NISPERO*) s->pinfo;

	if ((rc = RPCConnect(ctx)) != S_OK) {
		/*RPCDisconnect(ctx);*/
		return(rc);
	}

	/* Verify if [const char *host] controls ctx->node */	
	if (!strcmp(ctx->node, host)){
		syslog(LOG_WARNING, _("%s: %s (%s) [%s] "
				      "doesn't control host [%s]."), 
		       RPCid, ctx->idinfo, ctx->unitid, ctx->node, host);
		RPCDisconnect(ctx);
		return(S_BADHOST);
	}

	switch(request) {
		case ST_POWERON:
			rc = RPCOn(ctx, host);
			break;
		case ST_POWEROFF:
			rc = RPCOff(ctx, host);
			break;
		case ST_GENERIC_RESET:
			rc = RPCReset(ctx, host);
			break;
		default:
			rc = S_INVAL;
		break;
	}

	lorc = RPCDisconnect(ctx);

	return(rc != S_OK ? rc : lorc);
}

/*
 *	Parse the information in the given configuration file,
 *	and stash it away...
 */
static int
multip_nispero_set_config_file(Stonith* s, const char * configname)
{
	FILE *	cfgfile;

	char	cfgline[256];
	struct M_NISPERO*	ctx;
	if (gbl_debug) syslog(LOG_DEBUG, "%s: calling multip_nispero_set_config_file", RPCid);
	if (!ISM_NISPERO(s)) {
		syslog(LOG_ERR, "%s: invalid argument to RPC_set_configfile", RPCid);
		return(S_OOPS);
	}
	ctx = (struct M_NISPERO*) s->pinfo;

	if ((cfgfile = fopen(configname, "r")) == NULL)  {
		syslog(LOG_ERR, _("%s: cannot open %s"), RPCid, configname);
		return(S_BADCONFIG);
	}

	while (fgets(cfgline, sizeof(cfgline), cfgfile) != NULL){

		RPC_parse_config_info(ctx, cfgline);
	}
	return(S_BADCONFIG);
}

/*
 *	multip_nispero_set_config_info - API entry point to process one line of config info 
 *       for this particular device.
 *
 *      Parse the config information in the given string, and stash it away...
 *
 */
static int
multip_nispero_set_config_info(Stonith* s, const char * info)
{
	struct M_NISPERO* ctx;
	if (gbl_debug) syslog(LOG_DEBUG, "%s: calling multip_nispero_set_config_info", RPCid);
	if (!ISM_NISPERO(s)) {
		syslog(LOG_ERR, "%s: RPC_provide_config_info: invalid argument", RPCid);
		return(S_OOPS);
	}
	ctx = (struct M_NISPERO *)s->pinfo;

	return(RPC_parse_config_info(ctx, info));
}

/*
 * multip_nispero_getinfo - API entry point to retrieve something from the handle
 */
static const char *
multip_nispero_getinfo(Stonith * s, int reqtype)
{
	struct M_NISPERO *ctx;
	const char *ret;
	if (gbl_debug) syslog(LOG_DEBUG, "%s: calling multip_nispero_getinfo", RPCid);
	if (!ISM_NISPERO(s)) {
		syslog(LOG_ERR, "%s: RPC_idinfo: invalid argument", RPCid);
		return NULL;
	}
	/*
	 *	We look in the ST_TEXTDOMAIN catalog for our messages
	 */
	ctx = (struct M_NISPERO *)s->pinfo;

	switch (reqtype) {
		case ST_DEVICEID:
			ret = ctx->idinfo;
			break;

		case ST_CONF_INFO_SYNTAX:
			ret = _("<socket_path> <owner_node_hostname> <password> <reset time>\n"
			"All tokens are white-space delimited.");
			break;

		case ST_CONF_FILE_SYNTAX:
			ret = _("<socket_path> <owner_node_hostname> <password> <reset time>\n"
			"All tokens are white-space delimited.\n"
			"Blank lines and lines beginning with # are ignored");
			break;

		case ST_DEVICEDESCR:
			ret = _(DEVICE);
			break;

		case ST_DEVICEURL:
			ret = "http://www.multip.hu/";
			break;

		default:
			ret = NULL;
			break;
	}
	return ret;
}

/*
 * multip_nispero_destroy - API entry point to destroy a M_NISPERO Stonith object.
 */ 
static void
multip_nispero_destroy(Stonith *s)
{
	struct M_NISPERO* ctx;
	if (gbl_debug) syslog(LOG_DEBUG, "%s: multip_nispero_destroy", RPCid);
	if (!ISM_NISPERO(s)) {
		syslog(LOG_ERR, "%s: multip_nispero_destroy: invalid argument", RPCid);
		return;
	}
	ctx = (struct M_NISPERO *)s->pinfo;

	ctx->RPCid = NOTrpcid;

	/* RPCDisconnect(ctx);  <----- this is handled elsewhere */
	
	if (ctx->device != NULL) {
		FREE(ctx->device);
		ctx->device = NULL;
	}
	if (ctx->idinfo != NULL) {
		FREE(ctx->idinfo);
		ctx->idinfo = NULL;
	}
	if (ctx->unitid != NULL) {
		FREE(ctx->unitid);
		ctx->unitid = NULL;
	}
}

/* 
 * multip_nispero_new - API entry point called to create a new M_NISPERO Stonith device
 *          object. 
 */
static void *
multip_nispero_new(void)
{
	struct M_NISPERO*	ctx = MALLOCT(struct M_NISPERO);

	if (ctx == NULL) {
		syslog(LOG_ERR, "%s: out of memory", RPCid);
		return(NULL);
	}
	memset(ctx, 0, sizeof(*ctx));
	ctx->RPCid = RPCid;
	ctx->fd = -1;
	ctx->config = 0;
	ctx->device = NULL;
	ctx->node = NULL;
	ctx->idinfo = NULL;
	ctx->unitid = NULL;
	ctx->password = NULL;
	ctx->reset_time = RESET_INT; /* defaults to 10 secs */
	REPLSTR(ctx->idinfo, DEVICE);
	REPLSTR(ctx->unitid, "andres");
	REPLSTR(ctx->password,"");

	return((void *)ctx);
}
